/****************************************************************************/
/*
/*  BackBuffer, Copyright 1997-1999 by Longbow Digital Arts, Incorporated.
/*  Generic full screen and windowed linear frame buffer abstraction class,
/*  with support for simple dirty rectangle updating and true color or
/*  paletted display modes.  Written by Seumas McNally.  Used for all
/*  graphical display in the game "DX-Ball 2".
/*
/*  CreateDib, Copyright 1997-1999 by Michael P. Welch.
/*  Windows DIB wrapper.  Modifications by Seumas McNally.  Used for all
/*  graphical display in the screen saver "Particle Fire".
/*
/*  GLBuffer, Copyright 1998-1999 by Longbow Digital Arts, Incorporated.
/*  Work in progress wrapper for full screen and windowed OpenGL rendering.
/*  Used for OpenGL window handling in the game "Tread Marks".
/*
/****************************************************************************/
/*
/*  Intent:
/*  This source code is being released so that other programmers can make use
/*  of and possibly improve upon this simple, portable interface to linear
/*  memory display frame buffers.  The intent is to make the code as freely
/*  useable as possible, while still retaining some control for the original
/*  authors.  As such you are allowed to use the code under the following
/*  conditions:
/*
/*  1) If you use this source, portions of this source code, or a modification
/*  of same in a compiled program, you must include a credit line in your
/*  documentation, credits, or about window stating that the program makes use
/*  of BackBuffer, CreateDib, and/or GLBuffer code by Longbow Digital Arts
/*  and/or Michael P. Welch.
/*
/*  2) If you wish to publicly distribute modifications to this source code,
/*  this entire notice must remain intact, and your changes to the code must
/*  be reasonably documented in the Modification Log.  You are also asked to
/*  notify Longbow Digital Arts and Michael P. Welch about the public
/*  distribution of any changed versions.
/*  (http://www.LongbowDigitalArts.com/)
/*
/*  DISCLAIMER OF WARRANTY:
/*  This source code is provided "as is", without warranty of any kind, either
/*  express or implied, including but not limited to the warranties of
/*  merchantability and fitness for any particular purpose.
/*
/****************************************************************************/
/*
/*  Modification Log:
/*  - July 12th, 1999, public release.
/*
/****************************************************************************/

#include "GLBuffer.h"
#include <string.h>

#ifdef TREADMARKS
//Includes private logging functions.
#include <CfgParse.h>
#endif

#define NODDRAW

PFNGLACTIVETEXTUREARBPROC		glActiveTexture = NULL;
PFNGLMULTITEXCOORD2FARBPROC		glMultiTexCoord2f = NULL;
PFNGLMULTITEXCOORD2FVARBPROC	glMultiTexCoord2fv = NULL;
PFNGLCLIENTACTIVETEXTUREARBPROC		glClientActiveTexture = NULL;

PFNGLLOCKARRAYSEXTPROC		glLockArrays = NULL;
PFNGLUNLOCKARRAYSEXTPROC	glUnlockArrays = NULL;


GLBuffer::GLBuffer(){
#ifndef NODDRAW
	m_lpDD = NULL;
#endif
	m_hWnd = NULL;
	m_fullscreen = FALSE;
	hRC = NULL;
	hDC = NULL;
	m_WndProc = NULL;
	WinWidth = WinHeight = 1;
}
GLBuffer::~GLBuffer(){
	Destroy();
}

static GLBuffer *GLBufPtr = NULL;

LRESULT CALLBACK GLBufferProc(HWND hwnd, UINT iMsg, WPARAM wParam, LPARAM lParam){
	GLBuffer *glb = (GLBuffer*)GetWindowLong(hwnd, GWL_USERDATA);
	switch(iMsg){
		case WM_CREATE :
			glb = GLBufPtr;
			SetWindowLong(hwnd, GWL_USERDATA, (LONG)glb);
			break;
		case WM_ACTIVATEAPP :
			if(!glb) break;
			if(glb->m_wantfullscreen){// && glb->m_lpDD){
				if(wParam){	//Activating.
					ShowWindow(glb->m_hWnd, SW_SHOWNORMAL);//MAXIMIZE);
					MoveWindow(glb->m_hWnd, 0, 0, glb->m_w, glb->m_h, FALSE);
					glb->m_fullscreen = TRUE;
					wglMakeCurrent(glb->hDC, glb->hRC);
				}else{	//Deactivating.
					ShowWindow(glb->m_hWnd, SW_MINIMIZE);
					glb->m_fullscreen = FALSE;
					wglMakeCurrent(NULL, NULL);
				}
			}
			break;
		case WM_SYSCOMMAND :
			if(!glb) break;	//Disallow screensaver when we are playing fullscreen.
			if((wParam & 0xfff0) == SC_SCREENSAVE && glb->m_wantfullscreen) return 0;
			break;
		case WM_SIZE :
			if(!glb) break;
			if(glb->m_wantfullscreen == FALSE){
				RECT rc;
				GetClientRect(hwnd, &rc);
				glb->WinWidth = __max(rc.right, 1);
				glb->WinHeight = __max(rc.bottom, 1);
			}
			break;
		case WM_DESTROY :
			if(!glb) break;
			wglMakeCurrent(NULL, NULL);
			wglDeleteContext(glb->hRC);
			ReleaseDC(hwnd, glb->hDC);
			glb->Destroy(FALSE);
			break;
	}
	if(glb) return CallWindowProc((WNDPROC)glb->m_WndProc, hwnd, iMsg, wParam, lParam);
	return DefWindowProc(hwnd, iMsg, wParam, lParam);
}

int GLBuffer::CreateBuffer(int w, int h, int bpp, int fullscreen,
	const char *Name, HINSTANCE hInst, HWND *phWnd, LRESULT (CALLBACK *WProc)(HWND, UINT, WPARAM, LPARAM), DWORD Icon){
	char CName[1024];
	Destroy();

	strcpy(CName, Name);
	strcat(CName, "OpenGL");
	//
	m_wc.cbSize			= sizeof(m_wc);
	m_wc.style			= CS_HREDRAW | CS_VREDRAW | CS_OWNDC;
	m_wc.lpfnWndProc	= GLBufferProc;//WProc;
	m_wc.cbClsExtra		= 0;
	m_wc.cbWndExtra		= 0;
	m_wc.hInstance		= hInst;
	m_wc.hCursor		= LoadCursor(NULL, IDC_ARROW);
	m_wc.hbrBackground	= NULL;//(HBRUSH)GetStockObject(WHITE_BRUSH);
	m_wc.lpszMenuName	= NULL;
	m_wc.lpszClassName	= CName;
	if(Icon == NULL){
		m_wc.hIcon			= LoadIcon(NULL, IDI_APPLICATION);
		m_wc.hIconSm		= LoadIcon(NULL, IDI_APPLICATION);
	}else{
		m_wc.hIcon			= LoadIcon(hInst, MAKEINTRESOURCE(Icon));
		m_wc.hIconSm		= LoadIcon(hInst, MAKEINTRESOURCE(Icon));
	}
	RegisterClassEx(&m_wc);

	m_WndProc = WProc;
	m_wantfullscreen = fullscreen;
	m_fullscreen = FALSE;

	m_w = WinWidth = w;
	m_h = WinHeight = h;
	m_bpp = bpp;

	GLBufPtr = this;
	
	if(fullscreen){
		//
		DEVMODE dm;
		dm.dmSize = sizeof(dm);
		dm.dmBitsPerPel = bpp;
		dm.dmPelsWidth = w;
		dm.dmPelsHeight = h;
		dm.dmFields = DM_PELSWIDTH | DM_PELSHEIGHT | (bpp > 0 ? DM_BITSPERPEL : 0);
		if(DISP_CHANGE_SUCCESSFUL != ChangeDisplaySettings(&dm, CDS_FULLSCREEN)) return Destroy();
		m_fullscreen = TRUE;
		//
		*phWnd = m_hWnd = CreateWindowEx(WS_EX_TOPMOST, CName, Name,
			WS_POPUP | WS_MAXIMIZE | WS_CLIPSIBLINGS | WS_CLIPCHILDREN,
			CW_USEDEFAULT, CW_USEDEFAULT, w, h,
			NULL, NULL, hInst, NULL);
	}else{
		*phWnd = m_hWnd = CreateWindowEx(0, CName, Name,
			WS_OVERLAPPEDWINDOW | WS_CLIPSIBLINGS | WS_CLIPCHILDREN,
			CW_USEDEFAULT, CW_USEDEFAULT, w, h,
			NULL, NULL, hInst, NULL);
	}

	if(!m_hWnd) return Destroy();

	if(fullscreen){
		ShowWindow(m_hWnd, SW_SHOWNORMAL);//MAXIMIZE);
		ShowCursor(FALSE);
	}else{
		ShowWindow(m_hWnd, SW_SHOWNORMAL);
	}
	SetFocus(m_hWnd);

	if(fullscreen){
		BringWindowToTop(m_hWnd);
		SetForegroundWindow(m_hWnd);
	}

	PIXELFORMATDESCRIPTOR pfd;
	int pf;
	memset(&pfd, 0, sizeof(pfd)) ;
	pfd.nSize = sizeof(pfd);
	pfd.nVersion = 1;
	pfd.dwFlags = PFD_DOUBLEBUFFER | PFD_SUPPORT_OPENGL | PFD_DRAW_TO_WINDOW ;
	pfd.iPixelType = PFD_TYPE_RGBA;
	pfd.cColorBits = bpp;
	pfd.cDepthBits = 16;
	pfd.iLayerType = PFD_MAIN_PLANE;
	hDC = GetDC(m_hWnd);
	pf = ChoosePixelFormat(hDC, &pfd);
	if(!SetPixelFormat(hDC, pf, &pfd)) return Destroy();
	hRC = wglCreateContext(hDC);
	wglMakeCurrent(hDC, hRC);

#ifdef TREADMARKS
	//Output debug info.
	OutputDebugLog(CStr("\nGL_VENDOR:\n") + (char*)glGetString(GL_VENDOR) + "\n");
	OutputDebugLog(CStr("GL_RENDERER:\n") + (char*)glGetString(GL_RENDERER) + "\n");
	OutputDebugLog(CStr("GL_VERSION:\n") + (char*)glGetString(GL_VERSION) + "\n");
	OutputDebugLog(CStr("GL_EXTENSIONS:\n") + (char*)glGetString(GL_EXTENSIONS) + "\n\n");
#endif

	//Find MultiTexturing extension.
	glActiveTexture = NULL;
	glMultiTexCoord2f = NULL;
	glMultiTexCoord2fv = NULL;

	char *strExtensions = (char*)glGetString(GL_EXTENSIONS);
	if(strstr(strExtensions, "GL_ARB_multitexture") == 0){
#ifdef TREADMARKS
		OutputDebugLog("GL_ARB_multitexture NOT found.\n");
#endif
	}else{
#ifdef TREADMARKS
		OutputDebugLog("GL_ARB_multitexture detected.\n");
#endif
		glActiveTexture = (PFNGLACTIVETEXTUREARBPROC)wglGetProcAddress("glActiveTextureARB");
		glMultiTexCoord2f = (PFNGLMULTITEXCOORD2FARBPROC)wglGetProcAddress("glMultiTexCoord2fARB");
		glMultiTexCoord2fv = (PFNGLMULTITEXCOORD2FVARBPROC)wglGetProcAddress("glMultiTexCoord2fvARB");
		glClientActiveTexture = (PFNGLCLIENTACTIVETEXTUREARBPROC)wglGetProcAddress("glClientActiveTextureARB");
	}
	if(strstr(strExtensions, "GL_EXT_compiled_vertex_array") == 0){
#ifdef TREADMARKS
		OutputDebugLog("GL_EXT_compiled_vertex_array NOT found.\n");
#endif
	}else{
#ifdef TREADMARKS
		OutputDebugLog("GL_EXT_compiled_vertex_array detected.\n");
#endif
		glLockArrays = (PFNGLLOCKARRAYSEXTPROC)wglGetProcAddress("glLockArraysEXT");
		glUnlockArrays = (PFNGLUNLOCKARRAYSEXTPROC)wglGetProcAddress("glUnlockArraysEXT");
	}

	return TRUE;
}
int GLBuffer::Clear(){
	glClearColor(0, 0, 0, 0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	return TRUE;
}
int GLBuffer::Swap(){
	if(hDC){
		SwapBuffers(hDC);
		return TRUE;
	}
	return FALSE;
}

int GLBuffer::Destroy(int dwindow){
	//
	glActiveTexture = NULL;
	glMultiTexCoord2f = NULL;
	glMultiTexCoord2fv = NULL;

	glLockArrays = NULL;
	glUnlockArrays = NULL;
	//
#ifdef NODDRAW
	ChangeDisplaySettings(NULL, 0);
	if(m_fullscreen) ShowCursor(TRUE);
	m_fullscreen = FALSE;
#else
	if(m_lpDD){
		if(m_fullscreen){
			m_lpDD->RestoreDisplayMode();
			m_lpDD->SetCooperativeLevel(m_hWnd, DDSCL_NORMAL);
			m_fullscreen = FALSE;
		}
		m_lpDD->Release();
		m_lpDD = NULL;
	}
#endif
	if(m_hWnd && dwindow) DestroyWindow(m_hWnd);
	m_hWnd = NULL;

	m_fullscreen = FALSE;

	return FALSE;
}
