/****************************************************************************/
/*
/*  BackBuffer, Copyright 1997-1999 by Longbow Digital Arts, Incorporated.
/*  Generic full screen and windowed linear frame buffer abstraction class,
/*  with support for simple dirty rectangle updating and true color or
/*  paletted display modes.  Written by Seumas McNally.  Used for all
/*  graphical display in the game "DX-Ball 2".
/*
/*  CreateDib, Copyright 1997-1999 by Michael P. Welch.
/*  Windows DIB wrapper.  Modifications by Seumas McNally.  Used for all
/*  graphical display in the screen saver "Particle Fire".
/*
/*  GLBuffer, Copyright 1998-1999 by Longbow Digital Arts, Incorporated.
/*  Work in progress wrapper for full screen and windowed OpenGL rendering.
/*  Used for OpenGL window handling in the game "Tread Marks".
/*
/****************************************************************************/
/*
/*  Intent:
/*  This source code is being released so that other programmers can make use
/*  of and possibly improve upon this simple, portable interface to linear
/*  memory display frame buffers.  The intent is to make the code as freely
/*  useable as possible, while still retaining some control for the original
/*  authors.  As such you are allowed to use the code under the following
/*  conditions:
/*
/*  1) If you use this source, portions of this source code, or a modification
/*  of same in a compiled program, you must include a credit line in your
/*  documentation, credits, or about window stating that the program makes use
/*  of BackBuffer, CreateDib, and/or GLBuffer code by Longbow Digital Arts
/*  and/or Michael P. Welch.
/*
/*  2) If you wish to publicly distribute modifications to this source code,
/*  this entire notice must remain intact, and your changes to the code must
/*  be reasonably documented in the Modification Log.  You are also asked to
/*  notify Longbow Digital Arts and Michael P. Welch about the public
/*  distribution of any changed versions.
/*  (http://www.LongbowDigitalArts.com/)
/*
/*  DISCLAIMER OF WARRANTY:
/*  This source code is provided "as is", without warranty of any kind, either
/*  express or implied, including but not limited to the warranties of
/*  merchantability and fitness for any particular purpose.
/*
/****************************************************************************/
/*
/*  Modification Log:
/*  - July 12th, 1999, public release.
/*
/****************************************************************************/


/*
	GLBuffer OpenGL buffer window setup class.

	Paradigm shift from BackBuffer:  Due to the way OpenGL works
	under Windows, the window must be destroyed and recreated for
	screen mode changes, so now the Main Idle Loop must be God,
	and closing the window can not mean app death.  Put app death
	confirmation flag in WM_CLOSE or elsewhere, and have main loop
	post Quit Message.  Return 0 from WM_DESTROY.
*/

#ifndef GLBUFFER_H
#define GLBUFFER_H

#include <windows.h>
#include <gl/gl.h>
#include <gl/glu.h>
//#include <ddraw.h>

class GLBuffer{
friend LRESULT CALLBACK GLBufferProc(HWND hwnd, UINT iMsg, WPARAM wParam, LPARAM lParam);
public:
	GLBuffer();
	~GLBuffer();
	int CreateBuffer(int w, int h, int bpp, int fullscreen,
		const char *Name, HINSTANCE hInst, HWND *phWnd, LRESULT (CALLBACK *WProc)(HWND, UINT, WPARAM, LPARAM), DWORD Icon = NULL);
	int Destroy(int dwindow = FALSE);
	int Clear();
	int Swap();
private:
	BOOL	m_fullscreen, m_wantfullscreen;	//True if DirectDraw SetDisplayMode is in effect.
	WNDCLASSEX			m_wc;
	HWND				m_hWnd;
//	LPDIRECTDRAW		m_lpDD;	//Persistent member variables.
//	HRESULT			ddreturn;
	HGLRC hRC;
	HDC hDC;
	WNDPROC m_WndProc;
	int m_w, m_h, m_bpp;
public:
	int WinWidth, WinHeight;
};

//Extensions.

//ARB_multitexture
#define GL_ACTIVE_TEXTURE			0x84E0
#define GL_CLIENT_ACTIVE_TEXTURE	0x84E1
#define GL_MAX_TEXTURES_UNITS		0x84E2
#define GL_TEXTURE0					0x84C0
#define GL_TEXTURE1					0x84C1
#define GL_TEXTURE2					0x84C2
#define GL_TEXTURE3					0x84C3

typedef void (APIENTRY * PFNGLMULTITEXCOORD2FARBPROC) (GLenum target, GLfloat s, GLfloat t);
typedef void (APIENTRY * PFNGLMULTITEXCOORD2FVARBPROC) (GLenum target, const GLfloat *v);
typedef void (APIENTRY * PFNGLACTIVETEXTUREARBPROC) (GLenum target);
typedef void (APIENTRY * PFNGLCLIENTACTIVETEXTUREARBPROC) (GLenum target);

extern PFNGLACTIVETEXTUREARBPROC	glActiveTexture;
extern PFNGLMULTITEXCOORD2FARBPROC	glMultiTexCoord2f;
extern PFNGLMULTITEXCOORD2FVARBPROC	glMultiTexCoord2fv;
extern PFNGLCLIENTACTIVETEXTUREARBPROC	glClientActiveTexture;

//EXT_compiled_vertex_array

typedef void (APIENTRY * PFNGLLOCKARRAYSEXTPROC) (GLint first, GLuint count);
typedef void (APIENTRY * PFNGLUNLOCKARRAYSEXTPROC) (void);

extern PFNGLLOCKARRAYSEXTPROC	glLockArrays;
extern PFNGLUNLOCKARRAYSEXTPROC	glUnlockArrays;


#endif