//
// Simple Windows application demonstrating use of BackBuffer
// Written by Seumas McNally.  Placed into the Public Domain.
//

#include <BackBuffer.h>

//It is recommended that you place the BackBuffer files into a
//common source code directory and add that location as an
//Include File Path in your development environment, thus the
//#include <> instead of #include "".

//Press F8 to toggle full screen and windowed modes.

BackBuffer buf;
TrueColorFormat tcf;
PALETTEENTRY pe[256];

int WIDTH = 640;
int HEIGHT = 480;
int BPP = 8;
int FullScreen = FALSE;
int IsMinimized = 0;
int ActiveApp = 0;
int SwitchMode = 0;
int MouseX = 0, MouseY = 0;

HWND hWnd;

void DoFrame();
LRESULT CALLBACK WndProc(HWND hwnd, UINT iMsg, WPARAM wParam, LPARAM lParam);

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
				   PSTR szCmdLine, int iCmdShow){
	MSG	msg;

	if(!buf.InitWindow(WIDTH, HEIGHT, "BackBuffer Example", hInstance, &hWnd, WndProc, NULL)) return FALSE;
	if(!buf.SetBufferMode(WIDTH, HEIGHT, BPP, FullScreen, &tcf)) return FALSE;

	for(int i = 0; i < 256; i++){
		pe[i].peRed = __min(i * 2, 255);
		pe[i].peGreen = i;
		pe[i].peBlue = i / 4;
	}
	buf.SetPalette(pe);

	while(1){
		if(PeekMessage(&msg, NULL, 0, 0, PM_NOREMOVE)){
			if(!GetMessage(&msg, NULL, 0, 0)){
				return msg.wParam;
			}
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}else{
			if(IsMinimized == FALSE && (ActiveApp || !FullScreen)){
				DoFrame();
			}else{
				WaitMessage();
			}
		}
	}
}

void DoFrame(){
	if(SwitchMode){
		buf.SetBufferMode(WIDTH, HEIGHT, BPP, FullScreen, &tcf);
		buf.SetWindowSize(WIDTH, HEIGHT);
		buf.SetPalette(NULL);
		buf.RealizePalette();
		SwitchMode = FALSE;
	}
	//Per-frame stuff.
	BufferDesc bd;
	if(buf.Lock(&bd)){
		int x, y;
		unsigned char *p;
		//Set some random pixels.
		for(y = 1; y < HEIGHT - 1; y++){
			p = ((unsigned char*)bd.data + y * bd.pitch + rand() % (bd.width - 4) + 2);
			for(x = -1; x < 2; x++){
				p[x - bd.pitch] = 250;
				p[x] = 250;
				p[x + bd.pitch] = 250;
			}
		}
		//Burn fade buffer.
		for(y = 1; y < HEIGHT - 1; y++){
			p = (unsigned char*)bd.data + y * bd.pitch + 1;
			for(x = 1; x < WIDTH - 1; x++){
				int t = (((int)p[-1] + (int)p[0] + (int)p[1] + (int)p[bd.pitch]) >> 2) - 1;
				p[0] = t < 0 ? 0 : t;
				p++;
			}
		}
		buf.Unlock();
	}
	//Call UpdateFrontBuffer while UNLOCKED.
	buf.UpdateFrontBuffer(FullScreen ? NULL : UFB_STRETCH);
	//
}

LRESULT CALLBACK WndProc(HWND hwnd, UINT iMsg, WPARAM wParam, LPARAM lParam){
	PAINTSTRUCT ps;
	switch(iMsg){
		case WM_ACTIVATEAPP :
			ActiveApp = wParam;
			break;
		case WM_ACTIVATE :
			IsMinimized = HIWORD(wParam);
			break;
		case WM_CREATE :
			return 0;
		case WM_MOUSEMOVE :
			MouseX = LOWORD(lParam);
			MouseY = HIWORD(lParam);
			break;
		case WM_KEYDOWN :
			switch(wParam){
				case VK_ESCAPE :
					PostMessage(hwnd, WM_CLOSE, NULL, NULL);
					break;
				case VK_F8 :
					FullScreen = !FullScreen;
					SwitchMode = TRUE;
					break;
			}
			break;
		case WM_PAINT :
			BeginPaint(hwnd, &ps);
			buf.UpdateFrontBuffer(FullScreen ? NULL : UFB_STRETCH);
			EndPaint(hwnd, &ps);
			return 0;
		case WM_CLOSE :
			DestroyWindow(hwnd);
			return 0;
		case WM_DESTROY :
			hWnd = NULL;
			buf.Destroy();
			PostQuitMessage(0);
			return 0;
	}
	return DefWindowProc(hwnd, iMsg, wParam, lParam);
}

