/****************************************************************************/
/*
/*  BackBuffer, Copyright 1997-1999 by Longbow Digital Arts, Incorporated.
/*  Generic full screen and windowed linear frame buffer abstraction class,
/*  with support for simple dirty rectangle updating and true color or
/*  paletted display modes.  Written by Seumas McNally.  Used for all
/*  graphical display in the game "DX-Ball 2".
/*
/*  CreateDib, Copyright 1997-1999 by Michael P. Welch.
/*  Windows DIB wrapper.  Modifications by Seumas McNally.  Used for all
/*  graphical display in the screen saver "Particle Fire".
/*
/*  GLBuffer, Copyright 1998-1999 by Longbow Digital Arts, Incorporated.
/*  Work in progress wrapper for full screen and windowed OpenGL rendering.
/*  Used for OpenGL window handling in the game "Tread Marks".
/*
/****************************************************************************/
/*
/*  Intent:
/*  This source code is being released so that other programmers can make use
/*  of and possibly improve upon this simple, portable interface to linear
/*  memory display frame buffers.  The intent is to make the code as freely
/*  useable as possible, while still retaining some control for the original
/*  authors.  As such you are allowed to use the code under the following
/*  conditions:
/*
/*  1) If you use this source, portions of this source code, or a modification
/*  of same in a compiled program, you must include a credit line in your
/*  documentation, credits, or about window stating that the program makes use
/*  of BackBuffer, CreateDib, and/or GLBuffer code by Longbow Digital Arts
/*  and/or Michael P. Welch.
/*
/*  2) If you wish to publicly distribute modifications to this source code,
/*  this entire notice must remain intact, and your changes to the code must
/*  be reasonably documented in the Modification Log.  You are also asked to
/*  notify Longbow Digital Arts and Michael P. Welch about the public
/*  distribution of any changed versions.
/*  (http://www.LongbowDigitalArts.com/)
/*
/*  DISCLAIMER OF WARRANTY:
/*  This source code is provided "as is", without warranty of any kind, either
/*  express or implied, including but not limited to the warranties of
/*  merchantability and fitness for any particular purpose.
/*
/****************************************************************************/
/*
/*  Modification Log:
/*  - July 12th, 1999, public release.
/*
/****************************************************************************/

#include "cdib.h"

CreateDib::CreateDib(void)
{
	m_hWnd = NULL;
	m_sourceHdc = NULL;
	m_destHdc = NULL;
	memset(&m_dibInfo, 0, sizeof(CDIB_BITMAP));
	m_hBitmap = NULL;
	m_hOldBitmap = NULL;
	m_lpCDIBBits = NULL;
	memset(&m_logicalPalette, 0, sizeof(CDIBPALETTE));
	m_hPalette = NULL;
	m_hOldPalette = NULL;
	m_pitch = 0;
	m_bpp = 0;
}


CreateDib::~CreateDib(void)
{
	DeleteHBitmap();

	memset(&m_logicalPalette, 0, sizeof(CDIBPALETTE));
	if(m_hPalette)
	{
		DeleteObject(m_hPalette);
	}
	m_hPalette = NULL;
	m_hOldPalette = NULL;

	m_hWnd = NULL;

	if(m_sourceHdc)
	{
		DeleteDC(m_sourceHdc);
		m_sourceHdc = NULL;
	}

	if(m_destHdc)
	{
		ReleaseDC(m_hWnd, m_destHdc);
		m_destHdc = NULL;
	}
}

//True Color support added by Seumas McNally.
int CreateDib::CreateHBitmap(HWND hWnd, int width, int height, int bpp)
{
	DeleteHBitmap();

	bpp &= (~7);	//Wipe out lower three bits.
	if(bpp < 8 || bpp > 32 || width <= 0 || height <= 0) return FALSE;

	m_hWnd = hWnd;
	HDC hdc = GetDC(m_hWnd);
	m_bpp = bpp;
	m_pitch = (width * (bpp >>3) + 3) & (~3);	//Keeps it on dword boundary.

	// Fill in the BITMAPINFO structure with values
	m_dibInfo.bmiHead.biSize = sizeof(BITMAPINFOHEADER);
	m_dibInfo.bmiHead.biWidth = width;
	m_dibInfo.bmiHead.biHeight = -height;
	m_dibInfo.bmiHead.biPlanes = 1;
	m_dibInfo.bmiHead.biBitCount = bpp;
	m_dibInfo.bmiHead.biCompression = BI_RGB;
	m_dibInfo.bmiHead.biSizeImage = 0;
	m_dibInfo.bmiHead.biClrUsed = 0;
	m_dibInfo.bmiHead.biClrImportant = 0;

	m_hBitmap = CreateDIBSection(	hdc, 
								(BITMAPINFO*)&m_dibInfo, 
								DIB_PAL_COLORS, 
								&m_lpCDIBBits, 
								NULL, 
								NULL); 
	ReleaseDC(m_hWnd, hdc);
	if(m_hBitmap && m_lpCDIBBits) return TRUE;
	return FALSE;
}


void CreateDib::DeleteHBitmap(void)
{
	memset(&m_dibInfo, 0, sizeof(CDIB_BITMAP));
	if ( m_hBitmap ) 
	{
		HDC hdc = GetDC(m_hWnd);
		SelectObject( hdc, m_hOldBitmap );
		DeleteObject(m_hBitmap);
		ReleaseDC(m_hWnd, hdc);
	}
	m_hBitmap = NULL;
	m_hOldBitmap = NULL;
	m_lpCDIBBits = NULL;
	m_hWnd = NULL;
	m_pitch = 0;
	m_bpp = 0;
}


void CreateDib::SetPalette(PALETTEENTRY *palette)
{
	if(!m_hBitmap)
		return;

    m_logicalPalette.palVersion = 0x300;
    m_logicalPalette.palNumEntries = 256;

	if (m_hPalette != NULL)
	{
		DeleteObject(m_hPalette);
		m_hPalette = NULL;
	}

	UINT i;
    for( i = 0; i < 256; i++ ) 
	{
        m_dibInfo.rgb[i].rgbRed = 
			m_logicalPalette.aEntries[i].peRed = palette[i].peRed; 
        m_dibInfo.rgb[i].rgbGreen = 
            m_logicalPalette.aEntries[i].peGreen = palette[i].peGreen; 
        m_dibInfo.rgb[i].rgbBlue = 
            m_logicalPalette.aEntries[i].peBlue = palette[i].peBlue; 
        m_dibInfo.rgb[i].rgbReserved = 0;
        m_logicalPalette.aEntries[i].peFlags = 0;
    }
    m_hPalette = CreatePalette((LOGPALETTE*)&m_logicalPalette);

	HDC sHdc;
	sHdc = CreateCompatibleDC(NULL) ;
	m_hOldBitmap = (HBITMAP)SelectObject(sHdc, m_hBitmap);
	SetDIBColorTable( sHdc, 0, 256, m_dibInfo.rgb);
	SelectObject(sHdc, m_hOldBitmap);
	DeleteDC(sHdc);
}


void CreateDib::RealizePalette(void)
{
	if(!m_hPalette)
		return;
	
	HDC hdc = GetDC(m_hWnd);

	HPALETTE oldPalette = SelectPalette(hdc, m_hPalette, FALSE);	//New  (fixes bad resource leak)
	::RealizePalette(hdc);
	SelectPalette(hdc, oldPalette, TRUE);	//New
	ReleaseDC(m_hWnd,hdc);

//	InvalidateRect(m_hWnd,NULL,TRUE);
	//You may want to call same externally when you realize the palette, but not doing it here gives more control.  SM
}


BYTE* CreateDib::Lock(void)
{
	if(!m_hBitmap)
		return NULL;

	GdiFlush();

	if(!m_destHdc)
		m_destHdc = GetDC(m_hWnd);
	
	if(!m_sourceHdc)
		m_sourceHdc = CreateCompatibleDC(m_destHdc) ;

	m_hOldBitmap = (HBITMAP)SelectObject(m_sourceHdc, m_hBitmap);

	return (BYTE *)m_lpCDIBBits;
}


BOOL CreateDib::Blit(int destX, int destY, int sourceX, int sourceY, int sourceWidth, int sourceHeight)
{
	if(!m_hBitmap)
		return FALSE;

	BOOL result;
	HPALETTE oldPalette = SelectPalette(m_destHdc, m_hPalette, FALSE);	//New
	result = BitBlt(	m_destHdc, destX, destY, sourceWidth, sourceHeight,
						m_sourceHdc, sourceX, sourceY, SRCCOPY);
	SelectPalette(m_destHdc, oldPalette, TRUE);	//New

	return result;
}
//Function added by Seumas McNally.
BOOL CreateDib::StretchBlit(int destX, int destY, int destWidth, int destHeight, int sourceX, int sourceY, int sourceWidth, int sourceHeight)
{
	if(!m_hBitmap)
		return FALSE;

	BOOL result;
	HPALETTE oldPalette = SelectPalette(m_destHdc, m_hPalette, FALSE);	//New
	result = StretchBlt(	m_destHdc, destX, destY, destWidth, destHeight,
							m_sourceHdc, sourceX, sourceY, sourceWidth, sourceHeight, SRCCOPY);
	SelectPalette(m_destHdc, oldPalette, TRUE);	//New

	return result;
}
//Function added by Seumas McNally.
BOOL CreateDib::StretchBlitToWnd(){
	if(!m_hBitmap)
		return FALSE;
	BOOL result;
	RECT rc;
	GetClientRect(m_hWnd, &rc);
	HPALETTE oldPalette = SelectPalette(m_destHdc, m_hPalette, FALSE);	//New
	result = StretchBlt(	m_destHdc, 0, 0, rc.right, rc.bottom,
							m_sourceHdc, 0, 0, GetWidth(), GetHeight(), SRCCOPY);
	SelectPalette(m_destHdc, oldPalette, TRUE);	//New
	return result;
}
//Function added by Seumas McNally.
BOOL CreateDib::ScrollDib(int dx, int dy){
	if(!m_hBitmap) return FALSE;
	return ScrollDC(m_sourceHdc, dx, dy, NULL, NULL, NULL, NULL);
}

BOOL CreateDib::PaintBlit(HDC destHdc, int destX, int destY, int sourceX, int sourceY, int sourceWidth, int sourceHeight, int dW, int dH)
{
	if(!m_hBitmap)
		return FALSE;

	GdiFlush();

	// Lock
	HDC sourceHdc = CreateCompatibleDC(destHdc) ;
	m_hOldBitmap = (HBITMAP)SelectObject(sourceHdc, m_hBitmap);

	// Blit
	BOOL result;
	HPALETTE oldPalette = SelectPalette(m_destHdc, m_hPalette, FALSE);	//New
	if(dW == 0 && dH == 0){
		result = BitBlt(destHdc, destX, destY, sourceWidth, sourceHeight,
						sourceHdc, sourceX, sourceY, SRCCOPY);
	}else{
		result = StretchBlt(destHdc, destX, destY, dW, dH,
							sourceHdc, sourceX, sourceY, sourceWidth, sourceHeight, SRCCOPY);
	}
	SelectPalette(m_destHdc, oldPalette, TRUE);	//New

	// Unlock
	SelectObject(sourceHdc, m_hOldBitmap);
	DeleteDC(sourceHdc);

	return result;
}


void CreateDib::Unlock(void)
{
	if(!m_hBitmap)
		return;

	SelectObject(m_sourceHdc, m_hOldBitmap);

	DeleteDC(m_sourceHdc);
	m_sourceHdc = NULL;
	ReleaseDC(m_hWnd, m_destHdc);
	m_destHdc = NULL;
}

BYTE* CreateDib::GetSurface(void){
	if(!m_hBitmap)
		return NULL;
	return (BYTE *)m_lpCDIBBits;
}
int CreateDib::GetWidth(void){
	if(!m_hBitmap)
		return 0;
	return m_dibInfo.bmiHead.biWidth;
}
int CreateDib::GetHeight(void){
	if(!m_hBitmap)
		return 0;
	return abs(m_dibInfo.bmiHead.biHeight);
}
//Function added by Seumas McNally.
int CreateDib::GetPitch(void){
	if(!m_hBitmap)
		return 0;
	return m_pitch;
}
//Function added by Seumas McNally.
int CreateDib::BPP(){
	if(!m_hBitmap)
		return 0;
	return m_bpp;
}

//Helper function added by Seumas McNally.
int SetClientSize(HWND hwnd, int w, int h){
	RECT rcw, rcc;
	if(hwnd){
		GetWindowRect(hwnd, &rcw);
		GetClientRect(hwnd, &rcc);
		MoveWindow(hwnd, rcw.left, rcw.top, rcw.right - rcw.left + (w - rcc.right), rcw.bottom - rcw.top + (h - rcc.bottom), TRUE);
		return TRUE;
	}
	return FALSE;
}

//Helper function added by Seumas McNally.
int SetWindowPosition(HWND hwnd, int x, int y){
	RECT rcw;
	if(hwnd){
		GetWindowRect(hwnd, &rcw);
		MoveWindow(hwnd, x, y, rcw.right - rcw.left, rcw.bottom - rcw.top, TRUE);
		return TRUE;
	}
	return FALSE;
}
