/****************************************************************************/
/*
/*  BackBuffer, Copyright 1997-1999 by Longbow Digital Arts, Incorporated.
/*  Generic full screen and windowed linear frame buffer abstraction class,
/*  with support for simple dirty rectangle updating and true color or
/*  paletted display modes.  Written by Seumas McNally.  Used for all
/*  graphical display in the game "DX-Ball 2".
/*
/*  CreateDib, Copyright 1997-1999 by Michael P. Welch.
/*  Windows DIB wrapper.  Modifications by Seumas McNally.  Used for all
/*  graphical display in the screen saver "Particle Fire".
/*
/*  GLBuffer, Copyright 1998-1999 by Longbow Digital Arts, Incorporated.
/*  Work in progress wrapper for full screen and windowed OpenGL rendering.
/*  Used for OpenGL window handling in the game "Tread Marks".
/*
/****************************************************************************/
/*
/*  Intent:
/*  This source code is being released so that other programmers can make use
/*  of and possibly improve upon this simple, portable interface to linear
/*  memory display frame buffers.  The intent is to make the code as freely
/*  useable as possible, while still retaining some control for the original
/*  authors.  As such you are allowed to use the code under the following
/*  conditions:
/*
/*  1) If you use this source, portions of this source code, or a modification
/*  of same in a compiled program, you must include a credit line in your
/*  documentation, credits, or about window stating that the program makes use
/*  of BackBuffer, CreateDib, and/or GLBuffer code by Longbow Digital Arts
/*  and/or Michael P. Welch.
/*
/*  2) If you wish to publicly distribute modifications to this source code,
/*  this entire notice must remain intact, and your changes to the code must
/*  be reasonably documented in the Modification Log.  You are also asked to
/*  notify Longbow Digital Arts and Michael P. Welch about the public
/*  distribution of any changed versions.
/*  (http://www.LongbowDigitalArts.com/)
/*
/*  DISCLAIMER OF WARRANTY:
/*  This source code is provided "as is", without warranty of any kind, either
/*  express or implied, including but not limited to the warranties of
/*  merchantability and fitness for any particular purpose.
/*
/****************************************************************************/
/*
/*  Modification Log:
/*  - July 12th, 1999, public release.
/*
/****************************************************************************/

#ifndef CREATEDIB_H
#define CREATEDIB_H

#include <windows.h>
#include <windowsx.h>

class CreateDib
{
	public:

		// Constructor
		CreateDib(void);

		// Destructor
		~CreateDib(void);

		// Creation of hBitmap
		int CreateHBitmap(HWND hWnd, int width, int height, int bpp = 8);
		void DeleteHBitmap(void);

		// Palette Manipulation
		void SetPalette(PALETTEENTRY *palette);
		void RealizePalette(void);

		// Blitting interface 'Make sure to: 1) Lock surface, 2)Blit, 3)Unlock surface'
		BYTE* Lock(void);
		BOOL Blit(int destX, int destY, int sourceX, int sourceY, int sourceWidth, int sourceHeight);
		BOOL StretchBlit(int destX, int destY, int destWidth, int destHeight, int sourceX, int sourceY, int sourceWidth, int sourceHeight);
		BOOL StretchBlitToWnd();	//Blits to the dimensions of the attached window.
		BOOL ScrollDib(int dx, int dy);	//Wrap-around scrolls the contents of the bitmap.
		void Unlock(void);

		// Called from a WM_PAINT command (No locking required)
		BOOL PaintBlit(HDC destHdc, int destX, int destY, int sourceX, int sourceY, int sourceWidth, int sourceHeight, int dW = 0, int dH = 0);

		// Surface info
		BYTE* GetSurface(void);
		int GetWidth(void);
		int GetHeight(void);
		int GetPitch();
		BYTE *Data(){ return GetSurface(); };	//Simplified alternate interface.
		int Width(){ return GetWidth(); };
		int Height(){ return GetHeight(); };
		int Pitch(){ return GetPitch(); };
		int BPP();

		HBITMAP GetHBitmap(){ return m_hBitmap; };

	protected:

		//Added by Seumas McNally, July 1998.
		int m_pitch;
		int m_bpp;

		// Controlling window
		HWND m_hWnd;
		HDC m_sourceHdc;
		HDC m_destHdc;

		// BITMAPINFO
		typedef struct CDIB_BITMAP
		{
			BITMAPINFOHEADER bmiHead;
			RGBQUAD			 rgb[256];
		} CDIB_BITMAP;
		CDIB_BITMAP m_dibInfo;

		// Bitmap
		HBITMAP m_hBitmap;
		HBITMAP m_hOldBitmap;
		void far *m_lpCDIBBits;

		// Palette 
		typedef struct CDIBPALETTE { 
			WORD         palVersion; 
			WORD         palNumEntries; 
			PALETTEENTRY aEntries[256]; 
			} CDIBPALETTE; 	
		CDIBPALETTE m_logicalPalette;
		HPALETTE m_hPalette;
		HPALETTE m_hOldPalette;
};

int SetClientSize(HWND hwnd, int w, int h);
	//It sucks that Windows doesn't have a function
	//to do this, so I wrote one.  -Seumas
	//Sets Client Area size to that specified.
	//_Might_ have to be called twice in a row if you've
	//got a menu that's overlapping itself and will be
	//un-overlapped by size change.

	//It also sucks that Windows doesn't have a function
	//to merely set the position of a window without having
	//to specify its size too!
int SetWindowPosition(HWND hwnd, int x, int y);
	//Oops, I geuss it does, if you send the right flags to SetWindowPos().
	//Oh well.

#endif