//
// Simple Windows application demonstrating use of GLBuffer.
// Written by Seumas McNally.  Placed into the Public Domain.
//

#include <GLBuffer.h>

//It is recommended that you place the GLBuffer files into a
//common source code directory and add that location as an
//Include File Path in your development environment, thus the
//#include <> instead of #include "".

//Press F8 to toggle full screen and windowed modes.

#include <cdib.h>
//For the helper function SetClientSize() only.

GLBuffer glb;
HWND hWnd = NULL;
HINSTANCE hInst;
const char szAppName[] = "GLBuffer Demo";

int ActiveApp = FALSE;
int Quit = FALSE;
int UseFullScreen = FALSE;
int SwitchMode = FALSE;
int GLBWIDTH = 640;
int GLBHEIGHT = 480;
int GLBPP = 0;	//Use system BPP.

void DoFrame();

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, PSTR szCmdLine, int iCmdShow){
	MSG			msg;
	hInst = hInstance;

	while(1){
		if(PeekMessage(&msg, NULL, 0, 0, PM_NOREMOVE)){
			if(!GetMessage(&msg, NULL, 0, 0)){
				return msg.wParam;
			}
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}else{
			DoFrame();
		}
	}
}

LRESULT CALLBACK WndProc(HWND hwnd, UINT iMsg, WPARAM wParam, LPARAM lParam){
	switch(iMsg){
	case WM_ACTIVATEAPP :
		ActiveApp = wParam;
		break;
	case WM_KEYDOWN :
		switch(wParam){
		case VK_ESCAPE :
			PostMessage(hwnd, WM_CLOSE, 0, 0);
			break;
		case VK_F8 :
			SwitchMode = TRUE;
			UseFullScreen = !UseFullScreen;
			break;
		}
		break;
	case WM_CLOSE :
		Quit = TRUE;
		DestroyWindow(hwnd);
		return 0;
	case WM_DESTROY :
		hWnd = NULL;
		return 0;
	}
	return DefWindowProc(hwnd, iMsg, wParam, lParam);
}

void DoFrame(){
	if(Quit){
		PostQuitMessage(0);
		return;
	}
	if(SwitchMode || hWnd == NULL){
		if(hWnd) DestroyWindow(hWnd);
		if(glb.CreateBuffer(GLBWIDTH, GLBHEIGHT, GLBPP, UseFullScreen, szAppName, hInst, &hWnd, WndProc)){
			SwitchMode = FALSE;
			if(!UseFullScreen) SetClientSize(hWnd, GLBWIDTH, GLBHEIGHT);
			glb.Clear();
			glb.Swap();	//Give us a blank screen.
			return;
		}else{
			MessageBox(NULL, "Can't Open OpenGL Window!", "Error", MB_OK);
			if(UseFullScreen) UseFullScreen = FALSE;
			else Quit = TRUE;
			return;
		}
	}
	if(ActiveApp == FALSE){
		WaitMessage();
		return;
	}
	//
	glb.Clear();
	//
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0, 1, 0, 1, 0, 1);
	//
	glBegin(GL_TRIANGLES);
	glColor3f(1, 0, 0);
	glVertex3f(0.5, 0.1, -0.5);
	glColor3f(0, 1, 0);
	glVertex3f(0.1, 0.9, -0.5);
	glColor3f(0, 0, 1);
	glVertex3f(0.9, 0.9, -0.5);
	glEnd();
	//
	glb.Swap();
	//
}

